# -*- coding: utf-8 -*-
from ppms import ppms
from ppms.migration import SQLPacket, PacketCategory

try:
    from ppms.constants import DBMS_ORACLE, DBMS_MSSQL
    from ppms.ctg_aggregation import CostCalculation
    from ppms.calculations.budget.budget_calc_project import BudgetCalcProject
    from ppms.ctg_transfer import CtgTransfer
except ImportError:
    CtgTransfer = None
    

class MigrateCostsAndBudget(SQLPacket):
    '''
    Migrates costs and budget from the old cost type group model (five groups)
    into the new model with unlimited groups. Further initially calculates
    some new cost and budget columns, which didn't exist in the old model.
    This package does a cost and budget calculation of all projects and
    portfolios, which will need some time, depending on the amount of objects
    to calculate.
    '''
    version = 1
    done_after_success = True
    run_on_startup = True
    category = PacketCategory.CATEGORY_MANDATORY
    wi_number = 10938
        
    # Project migration statements
    mig_costs_pr_year = """insert into dt282
        (DI060470, -- pr_id
        DI060472,  -- year
        DI060473,  -- cost_type_group_id
        DI060476, DI060477, DI060478, DI060537, DI060538, DI060479, DI060480,
        DI060481, DI060482, DI060483,  -- object protection        
        DI060484, DI060485, DI060487, DI060488, -- ctg info
        DI060568, -- cost_planned_this
        DI060584, -- cost_act_this
        DI060583, -- cost_rem_this
        DI060586, -- cost_total_this
        DI060494, -- budget_submitted_cost
        DI060495, -- budget_approved_cost
        DI060503 -- cost_obligo
        )
        select 
            DI023119,   -- pr_id
            DI023062,   -- year
            '{ctg_id}', -- new cost_type_group_id
            DI023063,DI023064,DI023065,DI023066,DI023067,DI023068,DI023069,
            DI023070,DI023071,DI023072,  -- object protection
            DT280.DI060133, DT280.DI060132, DT280.DI060136, 1, -- ctg info
            {source_columns}
            from DT861
            join DT280 on ('{ctg_id}' = DT280.DI060121)
            """

    mig_eff_pr_year = """insert into dt282
        (DI060470, -- pr_id
        DI060472,  -- year
        DI060473,  -- cost_type_group_id
        DI060476, DI060477, DI060478, DI060537, DI060538, DI060479, DI060480,
        DI060481, DI060482, DI060483,  -- object protection        
        DI060484, DI060485, DI060487, DI060488, -- ctg info
        DI060585, -- effort_planned_this
        DI060587, -- effort_act_this
        DI060588, -- effort_rem_this
        DI060589, -- effort_total_this
        DI060513, -- budget_submitted_eff
        DI060506  -- budget_approved_eff
        )
        select 
            DI023119,   -- pr_id
            DI023062,   -- year
            '{ctg_id}', -- new cost_type_group_id
            DI023063,DI023064,DI023065,DI023066,DI023067,DI023068,DI023069,
            DI023070,DI023071,DI023072,  -- object protection
            DT280.DI060133, DT280.DI060132, DT280.DI060136, 2, -- ctg info
            {source_columns}
            from DT861
            join DT280 on ('{ctg_id}' = DT280.DI060121)
            """
                              
    mig_costs_pr = """insert into dt281
        (DI060139, -- pr_id
        DI060171, -- pr_id foreign key
        DI060140,  -- cost_type_group_id
        DI060161, DI060162, DI060163, DI060164, DI060165, DI060166, DI060167,
        DI060168, DI060169, DI060170,  -- object protection        
        DI060172, DI060173, DI060175, DI060176, -- ctg info
        DI060178 -- budget_proj_cost
        )
        select 
            DI001001,   -- pr_id
            DI001001,   -- pr_id
            '{ctg_id}', -- new cost_type_group_id
            DI001005,DI001006,DI001007,DI001008,DI001009,DI001010,DI000576,
            DI001011,DI001012,DI001013,  -- object protection
            DT280.DI060133, DT280.DI060132, DT280.DI060136, 1, -- ctg info
            {source_columns}
            from DT461
            join DT280 on ('{ctg_id}' = DT280.DI060121)
            """
    
    mig_effort_pr = """insert into dt281
        (DI060139, -- pr_id
        DI060171, -- pr_id foreign key
        DI060140,  -- cost_type_group_id
        DI060161, DI060162, DI060163, DI060164, DI060165, DI060166, DI060167,
        DI060168, DI060169, DI060170,  -- object protection        
        DI060172, DI060173, DI060175, DI060176, -- ctg info
        DI060190 -- budget_proj_eff
        )
        select 
            DI001001,   -- pr_id
            DI001001,   -- pr_id
            '{ctg_id}', -- new cost_type_group_id
            DI001005,DI001006,DI001007,DI001008,DI001009,DI001010,DI000576,
            DI001011,DI001012,DI001013,  -- object protection
            DT280.DI060133, DT280.DI060132, DT280.DI060136, 1, -- ctg info
            {source_columns}
            from DT461
            join DT280 on ('{ctg_id}' = DT280.DI060121)
            """
                            
    
    # Status report migration statements
    mig_report_costs_year = """insert into dt285
        (DI060386, -- pr_id
        DI060387,  -- report_id
        DI060388,  -- year
        DI060389,  -- cost_type_group_id
        DI060390, DI060391, DI060435, DI060436, DI060437,DI060392, DI060393,
        DI060394, DI060666, DI060395,  -- object protection        
        DI060396, DI060398, DI060399, DI060400, -- ctg info
        DI060408,  -- cost_planned_this
        DI060409,  -- cost_act_this
        DI060410,  -- cost_rem_this
        DI060411,  -- cost_total_this
        DI060413,  -- obligo_cost
        DI060406,  -- budget_approved_cost
        DI060405)   -- budget_submitted_cost
        select
            DI029230, -- pr_id
            DI029327, -- report_id
            DI029231, -- year
            '{ctg_id}', -- new cost_type_group_id
            DI029233,DI029234,DI029235,DI029236,DI029237,DI029238,DI029239,
            DI029240,DI029241,DI029242,  -- object protection
            DT280.DI060133, DT280.DI060132, DT280.DI060136, 1, -- ctg info
            {source_columns}
            from DT835
            join DT280 on ('{ctg_id}' = DT280.DI060121)
            """

    mig_report_costs_year_835_subpr = """insert into dt835
                        (DI029230, -- pr_id
                        DI029327,  -- report_id
                        DI029231,  -- year
                        DI029233, DI029234, DI029235, DI029236, DI029237,
                        DI029238, DI029239, DI029240, DI029241, DI029242)
                        values
                        ('{pr_id}','{report_id}',{year}, {dbms_timestamp},
                        {dbms_timestamp}, 7, 'MIGRATION', 'MIGRATION',
                         '{own_lic}',7,7,7,7)"""

    mig_report_costs_year_285_subpr = """insert into dt285
                (DI060386, -- pr_id
                DI060387,  -- report_id
                DI060388,  -- year
                DI060389,  -- cost_type_group_id
                DI060390, DI060391, DI060435, DI060436, DI060437,DI060392, DI060393,
                DI060394, DI060666, DI060395,  -- object protection
                DI060396, DI060398, DI060399, DI060400) -- ctg info
                values
                ('{pr_id}','{report_id}',{year},'{ctg_id}',
                {dbms_timestamp}, {dbms_timestamp}, 7, 'MIGRATION', 'MIGRATION',
                '{own_lic}',7,7,7,7,
                '{ctg_group}', '{ctg_name}', {ctg_sorting}, {ctg_type_code})"""

    mig_report_eff_year = """insert into dt285
        (DI060386, -- pr_id
        DI060387,  -- report_id
        DI060388,  -- year
        DI060389,  -- cost_type_group_id
        DI060390, DI060391, DI060435, DI060436, DI060437,DI060392, DI060393,
        DI060394, DI060666, DI060395,  -- object protection        
        DI060396, DI060398, DI060399, DI060400, -- ctg info
        DI060420, -- effort_planned_this
        DI060421, -- effort_act_this
        DI060422, -- effort_rem_this
        DI060423, -- effort_total_this
        DI060417, -- budget_submitted_eff
        DI060418 -- budget_approved_eff
        )
        select 
            DI029230, -- pr_id
            DI029327, -- report_id
            DI029231, -- year
            '{ctg_id}', -- new cost_type_group_id
            DI029233,DI029234,DI029235,DI029236,DI029237,DI029238,DI029239,
            DI029240,DI029241,DI029242,  -- object protection
            DT280.DI060133, DT280.DI060132, DT280.DI060136, 2, -- ctg info
            {source_columns}
            from DT835
            join DT280 on ('{ctg_id}' = DT280.DI060121)
            """
                              
    mig_report_costs_pr = """insert into dt284
        (DI060297, -- pr_id
        DI060300,  -- report_id
        DI060299,  -- cost_type_group_id
        DI060302, DI060303, DI060304, DI060305, DI060306, DI060307, DI060308,
        DI060309, DI060310, DI060311, -- object protection
        DI060312, DI060313, DI060315, DI060316, -- ctg info
        DI060318, -- budget_proj_cost
        DI060598, -- cost_planned_total
        DI060599, -- cost_act_total
        DI060600, -- cost_rem_total
        DI060601, -- cost_total_total
        DI060331 -- obligo_cost
        )
        select 
            DI008413, -- pr_id
            DI008416, -- report_id
            '{ctg_id}', -- new cost_type_group id
            DI008420, DI008421, DI008422, DI008423, DI008424, DI008425, DI008426,
            DI008427, DI008428, DI008429, -- object protection
            DT280.DI060133, DT280.DI060132, DT280.DI060136, 1, -- ctg info
            {source_columns}
            FROM DT823
            join DT280 on ('{ctg_id}' = DT280.DI060121)
            """

    calc_req_appr_budgets_costs = """update dt284 set
                        (DI060323,
                         DI060324)
                        = (select
                           {source_columns}
                           from DT835
                           where DT835.DI029230 = DT284.DI060297
                             and DT835.DI029327 = DT284.DI060300)
                        where DT284.DI060299 = '{ctg_id}'"""

    calc_req_appr_budgets_effort = """update dt284 set
                        (DI060335,
                         DI060336)
                        = (select
                           {source_columns}
                           from DT835
                           where DT835.DI029230 = DT284.DI060297
                             and DT835.DI029327 = DT284.DI060300)
                        where DT284.DI060299 = '{ctg_id}'"""

    calc_req_appr_budgets_costs_mssql = """update dt284 set
        DI060323 = year_sum.requested_budget,
        DI060324 = year_sum.approved_budget
        from (select DI029230,
                     DI029327,
                     {source_columns}
              from DT835
              group by DI029230, DI029327
              ) year_sum
        where DT284.DI060299 = '{ctg_id}'
        and year_sum.DI029230 = DT284.DI060297
        and year_sum.DI029327 = DT284.DI060300"""

    calc_req_appr_budgets_effort_mssql = """update dt284 set
        DI060335 = year_sum.budget_submitted_eff,
        DI060336 = year_sum.budget_approved_eff
        from (select DI029230,
                     DI029327,
                     {source_columns}
              from DT835
              group by DI029230, DI029327
              ) year_sum
        where DT284.DI060299 = '{ctg_id}'
        and year_sum.DI029230 = DT284.DI060297
        and year_sum.DI029327 = DT284.DI060300"""

    calc_req_appr_budgets_totals_costs = """
                        update dt284
                        set (DI061257,
                             DI061258)
                        = (select sum(nvl(DI060323,0)) budget_submitted_cost,
                                  sum(nvl(DI060324,0)) budget_approved_cost
                           from DT284 pr_report
                           where pr_report.DI060299=DT284.DI060299
                             and pr_report.DI060300='{report_id}'
                             and pr_report.DI060297 in {sub_pr_list}
                           )
                       where DI060297='{pr_id}'
                       and DI060300='{report_id}'"""

    calc_req_appr_budgets_totals_effort = """
                        update dt284
                        set (DI061260,
                             DI061261)
                        = (select sum(nvl(DI060335,0)) budget_submitted_eff,
                                  sum(nvl(DI060336,0)) budget_approved_eff
                           from DT284 pr_report
                           where pr_report.DI060299=DT284.DI060299
                             and pr_report.DI060300='{report_id}'
                             and pr_report.DI060297 in {sub_pr_list}
                           )
                       where DI060297='{pr_id}'
                       and DI060300='{report_id}'"""

    calc_req_appr_budgets_totals_costs_mssql = """update dt284
                    set DI061257 = sub_pr_sums.budget_submitted_cost,
                        DI061258 = sub_pr_sums.budget_approved_cost
                    from (select DI060299,
                                 sum(isnull(DI060323,0)) budget_submitted_cost,
                                 sum(isnull(DI060324,0)) budget_approved_cost
                          from DT284 pr_report
                          where pr_report.DI060297 in {sub_pr_list}
                                and DI060300='{report_id}'
                          group by DI060299
                          ) sub_pr_sums
                    where dt284.DI060300='{report_id}'
                      and sub_pr_sums.DI060299=DT284.DI060299
                      and dt284.DI060297='{pr_id}'"""

    calc_req_appr_budgets_totals_effort_mssql = """update dt284
                    set DI061260 = sub_pr_sums.budget_submitted_eff,
                        DI061261 = sub_pr_sums.budget_approved_eff
                    from (select DI060299,
                                 sum(isnull(DI060335,0)) budget_submitted_eff,
                                 sum(isnull(DI060336,0)) budget_approved_eff
                          from DT284 pr_report
                          where pr_report.DI060297 in {sub_pr_list}
                                and DI060300='{report_id}'
                          group by DI060299
                          ) sub_pr_sums
                    where dt284.DI060300='{report_id}'
                      and sub_pr_sums.DI060299=DT284.DI060299
                      and dt284.DI060297='{pr_id}'"""

    calc_cost_totals_report = """update dt284 set
                                (DI060327,
                                 DI060328,
                                 DI060326,
                                 DI060329)
                                = (select
                                    sum(nvl(DI060409,0)),
                                    sum(nvl(DI060410,0)),
                                    sum(nvl(DI060408,0)),
                                    sum(nvl(DI060411,0))
                                   from DT285
                                   where DT285.DI060386 = DT284.DI060297
                                     and DT285.DI060387 = DT284.DI060300
                                     and DT285.DI060389 = DT284.DI060299)"""

    calc_cost_totals_report_mssql = """update dt284 set
                                    DI060327 = year_sums.cost_act_this,
                                    DI060328 = year_sums.cost_rem_this,
                                    DI060326 = year_sums.cost_planned_this,
                                    DI060329 = year_sums.cost_total_this
                                from (select DI060386, DI060387, DI060389,
                                    sum(isnull(DI060409,0)) cost_act_this,
                                    sum(isnull(DI060410,0)) cost_rem_this,
                                    sum(isnull(DI060408,0)) cost_planned_this,
                                    sum(isnull(DI060411,0)) cost_total_this
                                    from DT285
                                    group by DI060386, DI060387, DI060389
                                    ) year_sums
                                where year_sums.DI060386 = DT284.DI060297
                                  and year_sums.DI060387 = DT284.DI060300
                                  and year_sums.DI060389 = DT284.DI060299"""

    mig_report_effort_pr = """insert into dt284
        (DI060297, -- pr_id
        DI060300,  -- report_id
        DI060299,  -- cost_type_group_id
        DI060302, DI060303, DI060304, DI060305, DI060306, DI060307, DI060308,
        DI060309, DI060310, DI060311, -- object protection
        DI060312, DI060313, DI060315, DI060316, -- ctg info
        DI060338, DI060339, DI060340, DI060341, -- effort planned/act/rem/total
        DI060602, DI060603, DI060604, DI060605) -- total effort planned/act/rem/total
        select 
            DI060386,   -- pr_id
            DI060387,   -- report_id
            DI060389,   -- cost_type_group_id
            max({dbms_timestamp}),max({dbms_timestamp}),max(7),max('{cur_user}'),max('{cur_user}')
            ,max('{usr_license}'),max(''),max(7),max(7),max(7), -- create object protection
            max(DT280.DI060133), max(DT280.DI060132), max(DT280.DI060136), max(2), -- ctg info
            sum(DI060420), -- effort_planned_this
            sum(DI060421), -- effort_act_this
            sum(DI060422), -- effort_rem_this
            sum(DI060423), -- effort_total_this
            sum(DI060610), -- effort_planned_total
            sum(DI060611), -- effort_act_total
            sum(DI060613), -- effort_rem_total
            sum(DI060612)  -- effort_total_total
            from DT285, DT280
            where DT285.DI060386 = '{pr_id}'
            and DT285.DI060387 = '{report_id}'
            and DT285.DI060400 = 2 -- ctg type (efforts)
            and DT285.DI060389 = DT280.DI060121
            group by DI060386,DI060387,DI060389"""

    mig_report_effort_pr_empty_rec = """insert into dt284
        (DI060297, -- pr_id
        DI060300,  -- report_id
        DI060299,  -- cost_type_group_id
        DI060302, DI060303, DI060304, DI060305, DI060306, DI060307, DI060308,
        DI060309, DI060310, DI060311, -- object protection
        DI060312, DI060313, DI060315, DI060316 -- ctg info
        )
        select
            '{pr_id}',       -- pr_id
            '{report_id}',   -- report_id
            '{ctg_id}',      -- cost_type_group_id
            {dbms_timestamp},{dbms_timestamp},7,'{cur_user}','{cur_user}'
            ,'{usr_license}','',7,7,7, -- object protection
            DT280.DI060133, DT280.DI060132, DT280.DI060136, 2 -- ctg info
            from DT280
            where DI060121='{ctg_id}'
            and not exists (select 1 from dt284
                              where DI060297='{pr_id}'
                              and DI060300='{report_id}'
                              and DI060299='{ctg_id}')
            """

    mig_report_current_budget = """
            update DT284
                set 
                DI060332 = 
                (select DI008615
                from DT823
                where DT823.DI008413 = DT284.DI060297
                  and DT823.DI008416 = DT284.DI060300)
                where DI060299='{ctg_id_a_intern}'"""
                
    mig_report_current_budget_mssql = """
            update DT284
                set DI060332 = source.effort_budget
                from
                (select DI008413 pr_id, 
                        DI008416 report_id, 
                        DI008615 effort_budget
                 from DT823) source
                where DT284.DI060299='{ctg_id_a_intern}'
                and source.pr_id = DT284.DI060297
                and source.report_id = DT284.DI060300
                """              

    calc_report_year_totals = """
        update dt285 set
           (DI060606, -- cost_planned_total
            DI060607, -- cost_act_total
            DI060608, -- cost_rem_total
            DI060609, -- cost_total_total
            DI060413, -- obligo_cost
            DI061264, -- budget_approved_cost_total
            DI061263, -- budget_submitted_cost_total
            DI060610, -- effort_planned_total
            DI060611, -- effort_act_total
            DI060613, -- effort_rem_total
            DI060612, -- effort_total_total
            DI061266, -- budget_submitted_eff_total
            DI061267) -- budget_approved_eff_total
        =
            (select
                nvl(sum(DI060408),0), -- cost_planned_this
                nvl(sum(DI060409),0), -- cost_act_this
                nvl(sum(DI060410),0), -- cost_rem_this
                nvl(sum(DI060411),0), -- cost_total_this
                nvl(sum(DI060413),0), -- obligo_cost
                nvl(sum(DI060406),0), -- budget_approved_cost
                nvl(sum(DI060405),0), -- budget_submitted_cost
                nvl(sum(DI060420),0), -- effort_planned_this
                nvl(sum(DI060421),0), -- effort_act_this
                nvl(sum(DI060422),0), -- effort_rem_this
                nvl(sum(DI060423),0), -- effort_total_this
                nvl(sum(DI060417),0), -- budget_submitted_eff
                nvl(sum(DI060418),0)  -- budget_approved_eff
            from DT285 status_yr
            where status_yr.DI060386 in {sub_pr_list}
            and status_yr.DI060389 = DT285.DI060389
            and status_yr.DI060387 = '{report_id}'
            and status_yr.DI060388 = DT285.DI060388)
        where DT285.DI060386 = '{pr_id}'
        and DT285.DI060387 = '{report_id}'"""
        
    calc_report_year_totals_mssql = """
        update dt285 set
            DI060606 = tpr_sum.cost_planned_total,
            DI060607 = tpr_sum.cost_act_total,
            DI060608 = tpr_sum.cost_rem_total,
            DI060609 = tpr_sum.cost_total_total,
            DI060413 = tpr_sum.obligo_cost,
            DI061264 = tpr_sum.budget_approved_cost_total,
            DI061263 = tpr_sum.budget_submitted_cost_total,
            DI060610 = tpr_sum.effort_planned_total,
            DI060611 = tpr_sum.effort_act_total,
            DI060613 = tpr_sum.effort_rem_total,
            DI060612 = tpr_sum.effort_total_total,
            DI061266 = tpr_sum.budget_submitted_eff_total,
            DI061267 = tpr_sum.budget_approved_eff_total
        from
            (select
                DI060389, DI060388,
                isnull(sum(DI060408),0) cost_planned_total,
                isnull(sum(DI060409),0) cost_act_total,
                isnull(sum(DI060410),0) cost_rem_total,
                isnull(sum(DI060411),0) cost_total_total,
                isnull(sum(DI060413),0) obligo_cost,
                isnull(sum(DI060406),0) budget_approved_cost_total,
                isnull(sum(DI060405),0) budget_submitted_cost_total,
                isnull(sum(DI060420),0) effort_planned_total,
                isnull(sum(DI060421),0) effort_act_total,
                isnull(sum(DI060422),0) effort_rem_total,
                isnull(sum(DI060423),0) effort_total_total,
                isnull(sum(DI060417),0) budget_submitted_eff_total,
                isnull(sum(DI060418),0) budget_approved_eff_total
            from DT285 status_yr
            where status_yr.DI060387 = '{report_id}'
            and status_yr.DI060386 in {sub_pr_list}
            group by DI060389, DI060388
            ) tpr_sum
        where DT285.DI060386 = '{pr_id}'
        and tpr_sum.DI060389 = DT285.DI060389
        and DT285.DI060387 = '{report_id}'
        and tpr_sum.DI060388 = DT285.DI060388
        """
            
    calc_budget_total_report = """update dt284       
        set (DI060322, DI060334) = (select 
                            nvl(sum(DI060318),0),  -- budget_proj_cost
                            nvl(sum(DI060332),0) -- budget_proj_eff
                            from DT284 pr
                            where pr.DI060297 in {sub_pr_list}
                            and pr.DI060300 = '{report_id}'
                            and pr.DI060299 = DT284.DI060299)
        where DI060297 = '{pr_id}'
          and DI060300 = '{report_id}'"""
        
    calc_budget_total_report_mssql = """update dt284       
                        set DI060322 = source.budget_proj_cost, 
                            DI060334 = source.budget_proj_eff
                            from
                            (select DI060299,
                                    isnull(sum(DI060318),0) budget_proj_cost,
                                    isnull(sum(DI060332),0) budget_proj_eff
                             from DT284 pr
                             where pr.DI060297 in {sub_pr_list}
                             and pr.DI060300 = '{report_id}'
                             group by DI060299
                            ) source
                        where dt284.DI060300 = '{report_id}'
                        and source.DI060299 = DT284.DI060299
                        and DI060297 = '{pr_id}'"""

    calc_budget_subproj_report= """update dt284
        set (DI060319, DI060333) = (select
                            nvl(sum(DI060318),0), -- budget_proj_cost
                            nvl(sum(DI060332),0) -- budget_proj_eff
                            from DT284 pr
                            where pr.DI060297 in {sub_pr_list}
                            and pr.DI060300 = '{report_id}'
                            and pr.DI060299 = DT284.DI060299)
        where DI060297 = '{pr_id}'
        and DI060300 = '{report_id}'"""

    calc_budget_subproj_report_mssql= """update dt284
                        set DI060319 = source.budget_proj_cost,
                            DI060333 = source.budget_proj_eff
                            from
                            (select DI060299,
                                isnull(sum(DI060318),0) budget_proj_cost,
                                isnull(sum(DI060332),0) budget_proj_eff
                            from DT284 pr
                            where pr.DI060297 in {sub_pr_list}
                            and pr.DI060300 = '{report_id}'
                            group by DI060299
                            ) source
                        where DI060300 = '{report_id}'
                        and source.DI060299 = DT284.DI060299
                        and DI060297 = '{pr_id}'"""

    calc_bl_budget_report_year = """
        update dt285
        set (DI060419,
             DI060407,
             DI061268,
             DI061265)
        = (select DI060423, DI060411,
                  DI060612, DI060609
           from DT285 active_bl
           where active_bl.DI060386 = DT285.DI060386
             and active_bl.DI060389 = DT285.DI060389
             and active_bl.DI060388 = DT285.DI060388
             and active_bl.DI060387 = (select DI041716 -- active baseline
                                          from DT823
                                          where DI008413 = (select DI008578 -- main pr_id
                                                            from DT823
                                                            where DI008413=dt285.DI060386
                                                              and DI008416=dt285.DI060387)
                                          and DI008416=dt285.DI060387)
            )"""

    calc_bl_budget_report_year_mssql = """
        update dt285
        set DI060419=bl_effort,
            DI060407=bl_costs,
            DI061268=bl_effort_total,
            DI061265=bl_costs_total
        from (select DI060386, DI060388, DI060389, DI060387,
                max(DI060423) bl_effort,
                max(DI060411) bl_costs,
                max(DI060612) bl_effort_total,
                max(DI060609) bl_costs_total
                from DT285 active_bl
              group by DI060386, DI060388, DI060389, DI060387
              ) baseline_values
        where baseline_values.DI060386 = DT285.DI060386
          and baseline_values.DI060388 = DT285.DI060388
          and baseline_values.DI060389 = DT285.DI060389
          and baseline_values.DI060387 = (select DI041716 -- active baseline
                                          from DT823
                                          where DI008413 = (select DI008578 -- main pr_id
                                                            from DT823 main_pr
                                                            where DI008413=dt285.DI060386
                                                              and DI008416=dt285.DI060387)
                                          and DI008416=dt285.DI060387)"""

    calc_bl_budget_report_project = """
        update dt284
        set (DI060337,  -- baseline_eff
             DI060325,  -- baseline_cost
             DI061262,  -- baseline_eff_total
             DI061259   -- baseline_cost_total
        )
        = (select DI060341, -- effort_total_this
                  DI060329, -- cost_total_this
                  DI060605, -- effort_total_total
                  DI060601  -- cost_total_total
           from DT284 active_bl
           where active_bl.DI060297 = DT284.DI060297
             and active_bl.DI060299 = DT284.DI060299
             and active_bl.DI060300 = (select DI041716 -- active baseline
                                          from DT823
                                          where DI008413 = (select DI008578 -- main pr_id
                                                            from DT823
                                                            where DI008413=DT284.DI060297
                                                              and DI008416=DT284.DI060300)
                                          and DI008416=dt284.DI060300)
            )"""

    calc_bl_budget_report_project_mssql = """update dt284
        set DI060337 = baseline_eff,
            DI060325 = baseline_cost,
            DI061262 = baseline_eff_total,
            DI061259 = baseline_cost_total
        from (select DI060297, DI060299, DI060300,
                     max(DI060341) baseline_eff,
                     max(DI060329) baseline_cost,
                     max(DI060605) baseline_eff_total,
                     max(DI060601) baseline_cost_total
           from DT284 active_bl
           group by DI060297, DI060299, DI060300
           ) baseline_values
       where baseline_values.DI060297 = DT284.DI060297
         and baseline_values.DI060299 = DT284.DI060299
         and baseline_values.DI060300 = (select DI041716 -- active baseline
                                          from DT823
                                          where DI008413 = (select DI008578 -- main pr_id
                                                            from DT823
                                                            where DI008413=DT284.DI060297
                                                              and DI008416=DT284.DI060300)
                                          and DI008416=dt284.DI060300)
       """

    create_report_pr = """insert into dt284
                (DI060297, -- program id
                DI060300,  -- report id
                DI060299,  -- ctg_id
                DI060312,  -- cost_type_group
                DI060313,  -- ctg_name
                DI060315,  -- ctg_sorting
                DI060316,  -- ctg_type_code
                DI060317,  -- ctg_type_name
                -- object protection
                DI060302,DI060303,DI060304,DI060305,DI060306,
                DI060307,DI060308,DI060309,DI060310,DI060311)
                values
                ('{obj_id}', '{report_id}', '{ctg_id}', '{cost_type_group}', 
                '{ctg_name}', {ctg_sorting}, '{ctg_type_code}', 
                '{ctg_type_name}', {dbms_timestamp}, {dbms_timestamp}, 7, 
                '{cur_user}', '{cur_user}', '{usr_license}',0,7,7,7)"""
                
    create_report_pr_year = """insert into dt285
                        (DI060386,  -- pr_id
                         DI060387,  -- report_id
                         DI060388,  -- year
                         DI060389,  -- cost_type_group_id
                         DI060390,  -- created_on
                         DI060391,  -- modified_on
                         DI060435,  -- rwdu
                         DI060436,  -- created_by
                         DI060437,  -- modified_by
                         DI060392,  -- own_lic
                         DI060393,  -- opc
                         DI060394,  -- owner
                         DI060666,  -- group
                         DI060395,  -- any
                         DI060396,  -- cost_type_group
                         DI060398,  -- ctg_name
                         DI060399,  -- ctg_sorting
                         DI060400,  -- ctg_type_code
                         DI060401   -- ctg_type_name
                        )
                        values ('{obj_id}','{report_id}',{year},'{ctg_id}',
                        {dbms_timestamp},{dbms_timestamp},7, '{cur_user}',
                        '{cur_user}', '{usr_license}', 7, 7, 0, 7, 
                        '{cost_type_group}', '{ctg_name}', {ctg_sorting}, 
                        '{ctg_type_code}', '{ctg_type_name}')"""
    
    set_effort_ctg_default = """update dt470 
                       set DI060521 = (case when di029478 = 1 
                                            then '{ctg_id_a_intern}'
                                            when di029478 = 2 
                                            then '{ctg_id_a_extern}'
                                            end)
                       where DI060517 is not Null"""
    
    set_ctg_to_kc_assignments = """
                                update dt470
                                     set DI060517 = (select DI060121 from dt280
                                     where dt280.DI060136 = dt470.di029478)"""
                                     
    set_ctg_to_kc_assignments_mssql = """
                                update dt470
                                     set DI060517 = source.ctg_id
                                     from
                                         (select DI060121 
                                             from dt280
                                         ) source
                                     where source.DI060136 = dt470.di029478"""
                                                  
    mig_pf_budget_int = """update DT283
                                set DI060362 =
                                (select 
                                DI040247
                                from DT859
                                where DT283.DI060343 = DT859.DI023041
                                and DT283.DI060345 = DT859.DI023053)
                                where DI060344 = '{ctg_id_int_k}'"""
                                
    mig_pf_budget_int_mssql = """update DT283
                                set DI060362 = source.budget_int
                                from
                                (select DI023041 portfolio_id,
                                        DI023053 period_id,
                                        DI040247 budget_int
                                 from DT859) source
                                where DI060344 = '{ctg_id_int_k}'
                                and DT283.DI060343 = source.portfolio_id
                                and DT283.DI060345 = source.period_id"""
                                
    mig_pf_budget_ext = """update DT283
                                set DI060362 =
                                (select 
                                DI040248
                                from DT859
                                where DT283.DI060343 = DT859.DI023041
                                and DT283.DI060345 = DT859.DI023053)
                                where DI060344 = '{ctg_id_ext_k}'"""
                                
    mig_pf_budget_ext_mssql = """update DT283
                                set DI060362 = source.budget_ext
                                from
                                (select DI023041 portfolio_id,
                                        DI023053 period_id,
                                        DI040248 budget_ext
                                 from DT859) source
                                where DI060344 = '{ctg_id_ext_k}'
                                and DT283.DI060343 = source.portfolio_id
                                and DT283.DI060345 = source.period_id"""

    mig_pf_budget_sach = """update DT283
                                set DI060362 =
                                (select 
                                DI040251
                                from DT859
                                where DT283.DI060343 = DT859.DI023041
                                and DT283.DI060345 = DT859.DI023053)
                                where DI060344 = '{ctg_id_sach_k}'"""
                                
    mig_pf_budget_sach_mssql = """update DT283
                                set DI060362 = source.budget_sach
                                from
                                (select DI023041 portfolio_id,
                                        DI023053 period_id,
                                        DI040251 budget_sach
                                from DT859) source
                                where DI060344 = '{ctg_id_sach_k}'
                                and DT283.DI060343 = source.portfolio_id
                                and DT283.DI060345 = source.period_id"""                                

    mig_pf_budget_inv = """update DT283
                                set DI060362 =
                                (select 
                                DI040249
                                from DT859
                                where DT283.DI060343 = DT859.DI023041
                                and DT283.DI060345 = DT859.DI023053)
                                where DI060344 = '{ctg_id_inv_k}'"""
                                
    mig_pf_budget_inv_mssql = """update DT283
                                set DI060362 = source.budget_invest
                                from
                                (select DI023041 portfolio_id,
                                        DI023053 period_id,
                                        DI040249 budget_invest
                                from DT859) source
                                where DI060344 = '{ctg_id_inv_k}'
                                and DT283.DI060343 = source.portfolio_id
                                and DT283.DI060345 = source.period_id"""                            

    mig_pf_budget_sonst = """update DT283
                                set DI060362 =
                                (select 
                                DI040250
                                from DT859
                                where DT283.DI060343 = DT859.DI023041
                                and DT283.DI060345 = DT859.DI023053)
                                where DI060344 = '{ctg_id_sonst_k}'"""
                                
    mig_pf_budget_sonst_mssql = """update DT283
                                set DI060362 = source.budget_sonst
                                from
                                (select DI023041 portfolio_id,
                                        DI023053 period_id,
                                        DI040250 budget_sonst
                                from DT859) source
                                where DI060344 = '{ctg_id_sonst_k}'
                                and DT283.DI060343 = source.portfolio_id
                                and DT283.DI060345 = source.period_id"""

    mig_pf_eff_budget_int = """update DT283
                                set DI060374 =
                                (select DI041721
                                from DT859
                                where DT283.DI060343 = DT859.DI023041
                                and DT283.DI060345 = DT859.DI023053)
                                where DI060344 = '{ctg_id_a_intern}'"""

    mig_pf_eff_budget_int_mssql  = """update DT283
                                    set DI060374 = source.budget_eff_int
                                    from
                                    (select DI023041 portfolio_id,
                                            DI023053 period_id,
                                            DI041721 budget_eff_int
                                    from DT859) source
                                    where DI060344 = '{ctg_id_a_intern}'
                                    and DT283.DI060343 = source.portfolio_id
                                    and DT283.DI060345 = source.period_id"""

    mig_pf_eff_budget_ext = """update DT283
                                set DI060374 =
                                (select DI041720
                                from DT859
                                where DT283.DI060343 = DT859.DI023041
                                and DT283.DI060345 = DT859.DI023053)
                                where DI060344 = '{ctg_id_a_extern}'"""

    mig_pf_eff_budget_ext_mssql  = """update DT283
                                    set DI060374 = source.budget_eff_ext
                                    from
                                    (select DI023041 portfolio_id,
                                            DI023053 period_id,
                                            DI041720 budget_eff_ext
                                    from DT859) source
                                    where DI060344 = '{ctg_id_a_extern}'
                                    and DT283.DI060343 = source.portfolio_id
                                    and DT283.DI060345 = source.period_id"""

    set_last_bl_costs_pr = """update dt281 set
            (di060200, -- bl_cost_this
             di060183, -- bl_cost_total
             di060188, -- bl_eff_this
             di060195) -- bl_eff_total
             = (select DI060329 cost_total_this,
                       DI060601 cost_total_total,
                       DI060341 effort_total_this,
                       DI060605 effort_total_total
               from DT284
               where DT284.DI060297 = DT281.DI060139 and
                     DT284.DI060299 = DT281.DI060140
                     and DT284.DI060300 = (
                              select DI041684 -- active baseline
                              from DT461
                              where DI001001 = (select DI001052 -- main pr_id
                                                from DT461
                                                where DI001001=DT284.DI060297)))
                              """

    set_last_bl_costs_pr_mssql = """update dt281 set
                di060200 = cost_total_this,
                di060183 = cost_total_total,
                di060188 = effort_total_this,
                di060195 = effort_total_total
                from (select DI060297 pr_id,
                             DI060299 cost_type_group,
                             DI060329 cost_total_this,
                             DI060601 cost_total_total,
                             DI060341 effort_total_this,
                             DI060605 effort_total_total
                    from DT284
                    where DT284.DI060300 = (select DI041684 -- active baseline
                                            from DT461
                                            where DI001001 = (select DI001052 -- main pr_id
                                                                from DT461
                                                                where DI001001=DT284.DI060297))
                    ) as report_vals
                where report_vals.pr_id = DT281.DI060139
                  and report_vals.cost_type_group = DT281.DI060140"""

    set_last_bl_costs_year = """update dt282 set
                (DI060496,
                 DI060502,
                 DI060512,
                 DI060507)
                 = (select DI060411, -- cost_total_this
                           DI060609, -- cost_total_total
                           DI060423, -- effort_total_this
                           DI060612  -- effort_total_total
                    from DT285
                    where DT285.DI060386=DT282.DI060470
                      and DT285.DI060388=DT282.DI060472
                      and DT285.DI060389=DT282.DI060473
                      and DT285.DI060387 = (select DI041684 -- active baseline
                                            from DT461
                                            where DI001001 = (select DI001052 -- main pr_id
                                                              from DT461
                                                              where DI001001=DT285.DI060386)))
                 """

    set_last_bl_costs_year_mssql = """update dt282 set
            DI060496 = cost_total_this,
            DI060502 = cost_total_total,
            DI060512 = effort_total_this,
            DI060507 = effort_total_total
            from (select
                    DI060386 pr_id,
                    DI060388 pr_year,
                    DI060389 cost_type_group,
                    DI060411 cost_total_this,
                    DI060609 cost_total_total,
                    DI060423 effort_total_this,
                    DI060612 effort_total_total
                    from DT285
                    where DT285.DI060387 = (select DI041684 -- active baseline
                                            from DT461
                                            where DI001001 = (select DI001052 -- main pr_id
                                                              from DT461
                                                              where DI001001=DT285.DI060386))
                ) as report_vals
            where report_vals.pr_id=DT282.DI060470
              and report_vals.pr_year=DT282.DI060472
              and report_vals.cost_type_group=DT282.DI060473"""

    set_last_report_year = """update dt282 set
        (DI061242, -- cost_total_active_sr_total
        DI061243,  -- effort_total_active_sr_total
        DI061244,  -- cost_total_active_sr_this
        DI061245,  -- effort_total_active_sr_this
        DI061367,  -- cost_plan_active_sr_total
        DI061368,  -- cost_act_active_sr_total
        DI061369,  -- cost_rem_active_sr_total
        DI061370,  -- budget_proj_cost_active_sr
        DI061371,  -- budget_proj_total_cost_active_sr
        DI061373,  -- budget_submitted_cost_active_sr_total
        DI061374,  -- budget_approved_cost_active_sr_total
        DI061375,  -- budget_baseline_cost_active_sr_total
        DI061377,  -- cost_obligo_active_sr
        DI061378,  -- effort_plan_active_sr_total
        DI061379,  -- effort_act_active_sr_total
        DI061380,  -- effort_rem_active_sr_total
        DI061381,  -- budget_proj_eff_active_sr
        DI061382,  -- budget_proj_total_eff_active_sr
        DI061383,  -- budget_submitted_eff_active_sr_total
        DI061384,  -- budget_approved_eff_active_sr_total
        DI061385)  -- budget_baseline_eff_active_sr_total
        =
        (select
        DI060609, -- cost_total_total
        DI060612, -- effort_total_total
        DI060411, -- cost_total_this
        DI060423, -- effort_total_this
        DI060606, -- cost_planned_total
        DI060607, -- cost_act_total
        DI060608, -- cost_rem_total
        DI060402, -- budget_proj_cost
        DI060404, -- budget_proj_total_cost
        DI061263, -- budget_submitted_cost_total
        DI061264, -- budget_approved_cost_total
        DI061265, -- budget_baseline_cost_total
        DI060413, -- obligo_cost
        DI060610, -- effort_planned_total
        DI060611, -- effort_act_total
        DI060613, -- effort_rem_total
        DI060414, -- budget_proj_eff
        DI060416, -- budget_proj_total_eff
        DI061266, -- budget_submitted_eff_total
        DI061267, -- budget_approved_eff_total
        DI061268  -- budget_baseline_eff_total
        from DT285
        where DT285.DI060386 = DT282.DI060470
          and DT285.DI060388 = DT282.DI060472
          and DT285.DI060389 = DT282.DI060473
          and DT285.DI060387 = (select DI041685 -- current report
                                from DT461
                                where DI001001=DT285.DI060386))
        """

    set_last_report_year_mssql = """update dt282 set
        DI061242 = source.cost_total_total,
        DI061243 = source.effort_total_total,
        DI061244 = source.cost_total_this,
        DI061245 = source.effort_total_this,
        DI061367 = source.cost_planned_total,
        DI061368 = source.cost_act_total,
        DI061369 = source.cost_rem_total,
        DI061370 = source.budget_proj_cost,
        DI061371 = source.budget_proj_total_cost,
        DI061373 = source.budget_submitted_cost_total,
        DI061374 = source.budget_approved_cost_total,
        DI061375 = source.budget_baseline_cost_total,
        DI061377 = source.obligo_cost,
        DI061378 = source.effort_planned_total,
        DI061379 = source.effort_act_total,
        DI061380 = source.effort_rem_total,
        DI061381 = source.budget_proj_eff,
        DI061382 = source.budget_proj_total_eff,
        DI061383 = source.budget_submitted_eff_total,
        DI061384 = source.budget_approved_eff_total,
        DI061385 = source.budget_baseline_eff_total
    from(
        select
            DI060386 pr_id,
            DI060388 pr_year,
            DI060389 cost_type_group,
            DI060609 as cost_total_total,
            DI060612 as effort_total_total,
            DI060411 as cost_total_this,
            DI060423 as effort_total_this,
            DI060606 as cost_planned_total,
            DI060607 as cost_act_total,
            DI060608 as cost_rem_total,
            DI060402 as budget_proj_cost,
            DI060404 as budget_proj_total_cost,
            DI061263 as budget_submitted_cost_total,
            DI061264 as budget_approved_cost_total,
            DI061265 as budget_baseline_cost_total,
            DI060413 as obligo_cost,
            DI060610 as effort_planned_total,
            DI060611 as effort_act_total,
            DI060613 as effort_rem_total,
            DI060414 as budget_proj_eff,
            DI060416 as budget_proj_total_eff,
            DI061266 as budget_submitted_eff_total,
            DI061267 as budget_approved_eff_total,
            DI061268 as budget_baseline_eff_total
        from DT285
        where DI060387 = (select DI041685 -- current report
                          from DT461
                          where DI001001=DT285.DI060386)
        ) source
    where DI060470=source.pr_id
      and DI060472=source.pr_year
      and DI060473=source.cost_type_group"""

    activate_pf_years = """insert into dt286
                 (DI061324, -- pr_id
                 DI061325, -- portfolio_id
                 DI061326, -- year
                 DI061328, -- created_on
                 DI061329, -- modified_on
                 DI061330, -- rwdu
                 DI061331, -- created_by
                 DI061332, -- modified_by
                 DI061333, -- own_lic
                 DI061334, -- opc
                 DI061335, -- owner
                 DI061336, -- group
                 DI061337, -- any
                 DI061338  -- active
                 ) select
                 DT858.DI023101, -- pr_id
                 DT858.DI023102, -- portfolio_id
                 DT861.DI023062, -- year
                 {dbms_timestamp},{dbms_timestamp},7,'{cur_user}','{cur_user}',
                 '{usr_license}','',7,7,7, -- create object protection
                 1 -- set year active
                 from dt858
                 join dt861 on (dt858.DI023101 = dt861.DI023119)
                 """

    copy_benefit_revenues = """update dt835
                                set BEN_PLAN_THIS=nvl(DI029263,0),
                                    BEN_ACH_THIS=nvl(DI029316,0),
                                    REV_ACT_THIS=nvl(DI029313,0),
                                    REV_REM_THIS=nvl(DI029314,0),
                                    REV_TOT_THIS=nvl(DI029315,0),
                                    REV_PLAN_THIS=nvl(DI029312,0)"""

    copy_benefit_revenues_mssql = """update dt835
                                set BEN_PLAN_THIS=isnull(DI029263,0),
                                    BEN_ACH_THIS=isnull(DI029316,0),
                                    REV_ACT_THIS=isnull(DI029313,0),
                                    REV_REM_THIS=isnull(DI029314,0),
                                    REV_TOT_THIS=isnull(DI029315,0),
                                    REV_PLAN_THIS=isnull(DI029312,0)"""

    calc_benefit_revenues = """update dt835
                        set (DI029263,
                            DI029316,
                            DI029313,
                            DI029314,
                            DI029315,
                            DI029312)
                        = (select sum(nvl(BEN_PLAN_THIS,0)),
                                  sum(nvl(BEN_ACH_THIS,0)),
                                  sum(nvl(REV_ACT_THIS,0)),
                                  sum(nvl(REV_REM_THIS,0)),
                                  sum(nvl(REV_TOT_THIS,0)),
                                  sum(nvl(REV_PLAN_THIS,0))
                           from DT835 struct_sum
                           where struct_sum.DI029230 in {sub_pr_list}
                             and struct_sum.DI029327='{report_id}'
                             and struct_sum.DI029231=DT835.DI029231)
                        where DT835.DI029230 = '{pr_id}'
                          and DT835.DI029327 = '{report_id}'"""

    calc_benefit_revenues_mssql = """update dt835
                        set DI029263 = struct_sum.BEN_PLAN_THIS,
                            DI029316 = struct_sum.BEN_ACH_THIS,
                            DI029313 = struct_sum.REV_ACT_THIS,
                            DI029314 = struct_sum.REV_REM_THIS,
                            DI029315 = struct_sum.REV_TOT_THIS,
                            DI029312 = struct_sum.REV_PLAN_THIS
                        from (select DI029231,
                                    sum(isnull(BEN_PLAN_THIS,0)) BEN_PLAN_THIS,
                                    sum(isnull(BEN_ACH_THIS,0)) BEN_ACH_THIS,
                                    sum(isnull(REV_ACT_THIS,0)) REV_ACT_THIS,
                                    sum(isnull(REV_REM_THIS,0)) REV_REM_THIS,
                                    sum(isnull(REV_TOT_THIS,0)) REV_TOT_THIS,
                                    sum(isnull(REV_PLAN_THIS,0)) REV_PLAN_THIS
                            from DT835
                            where DT835.DI029230 in {sub_pr_list}
                              and DT835.DI029327 = '{report_id}'
                            group by DI029231
                            ) struct_sum
                        where DT835.DI029327='{report_id}'
                          and DT835.DI029230='{pr_id}'
                          and struct_sum.DI029231=DT835.DI029231"""

    calc_resulting_benefit = """update DT835
                                set DI061418 = (case when DI029316 = 0
                                                  then DI029263
                                                  else DI029316
                                                  end),
                                    BEN_RES_THIS = (case when BEN_ACH_THIS = 0
                                                      then BEN_PLAN_THIS
                                                      else BEN_ACH_THIS
                                                      end)"""



    def fix(self, runner):
        # this replaces a dependency check (which can't be used as this runs
        # after the db migration itself) and makes sure below is only executed
        # once per system
        if self.new_dt280_already_filled():
            return self.success()

        self.set_global_setting_budget_calculation()
    
        self.dbms_name = ppms.db_get_dbms_name()
        self.dbms_timestamp = {DBMS_ORACLE: 'SYSTIMESTAMP', 
                               DBMS_MSSQL: 'GETDATE()'}[self.dbms_name]
        self.cur_user = ppms.uvar_get('@1')
        self.usr_license = ppms.uvar_get('@3')

        # get a dict of all projects and their subproject list per report
        self.all_report_pr_dict = self.get_sub_pr_dict()

        # Cost type group base data: migrate data from DT495 to DT280 
        # and within DT470
        ctgs_created = self.migrate_cost_type_groups()
        if not ctgs_created:
            return self.fail()

        # Project: migrate project budget data from DT461/861 to new 
        # tables DT281/282
        self.migrate_budget_years()
        self.migrate_budget_project()
        
        # Project status reports: migrate report budget data from DT835/832 
        # to new tables DT285/284
        self.migrate_report_years()
        self.migrate_report_project()

        # create records in DT281-DT285 for all cost type groups which were not
        # created within migration
        transfer = CtgTransfer()
        transfer_result = transfer.start_ctg_transfer()
        self.log('Result of project and portfolio cost type group transfer: ' \
                 '{transfer_result}'.format(**locals()))
         
        transfer_result = self.start_report_ctg_transfer()
        self.log('Result of report cost type group transfer: ' \
                 '{transfer_result}'.format(**locals()))

        # set last report/baseline values for projects
        self.set_last_report_values()

        # start a full budget calculation of all projects
        for pr_id in transfer.projects:
            pr_id = pr_id[0]
            BudgetCalcProject(pr_id).start_budget_calc()
            self.log('Budget calculation of project ' \
                     '{pr_id} done.'.format(**locals()))
        
        # start a full cost calculation of all projects and portfolios
        ctg_aggregation_full = CostCalculation(calc_inactive_objects=True)
        projects = ctg_aggregation_full.projects
        self.log('Starting cost calculation of all projects: {projects}'.format(**locals()))
        pr_load_info = ctg_aggregation_full.start_full_aggregation()
        self.log('Project cost calculation done. pr_load_info: {pr_load_info}'.format(**locals()))

        # migrate Nutzen+Erlöse in status reports
        self.migrate_benefit_and_revenues()

        # Portfolio migrations
        self.activate_pr_year_pf()

        # make sure all project values are migrated and portfolio years
        # are activated
        ppms.db_commit()

        self.migrate_portfolio_budget()

        portfolios = ctg_aggregation_full.portfolios
        self.log('Starting cost calculation of portfolios: ' \
                 '{portfolios}'.format(**locals()))
        ctg_aggregation_full.start_full_aggregation_pf(years=None)
        self.log('Portfolio cost calculation done.')

        return self.success()

    def get_sub_pr_dict(self):
        # create a dictionary of all projects and their subprojects as list
        all_reports = ppms.db_select("select DI008413, DI008416 from DT823")
        all_report_pr_dict = {}
        for pr_info in all_reports:
            pr_id = pr_info[0]
            report_id = pr_info[1]

            if pr_id not in all_report_pr_dict.keys():
                all_report_pr_dict[pr_id] = {}

            all_report_pr_dict[pr_id][report_id] = self.get_pr_tree_desc(pr_id,
                                                              report_id,
                                                              list_format='db')

        return all_report_pr_dict

    def set_last_report_values(self):
        # DT282: initialise values for "current report" / "baseline"
        # on project and project/year level for all existing projects.

        # last report values
        if self.DBMS == DBMS_ORACLE:
            mig_stmt = self.set_last_report_year.format(**locals())
        elif self.DBMS == DBMS_MSSQL:
            mig_stmt = self.set_last_report_year_mssql.format(**locals())
        self.modify_and_log(mig_stmt)

        # baseline values
        if self.DBMS == DBMS_ORACLE:
            self.modify_and_log(self.set_last_bl_costs_pr)
            self.modify_and_log(self.set_last_bl_costs_year)
        elif self.DBMS == DBMS_MSSQL:
            self.modify_and_log(self.set_last_bl_costs_pr_mssql)
            self.modify_and_log(self.set_last_bl_costs_year_mssql)

    def set_global_setting_budget_calculation(self):
        # Set p_project_budget_edit to 0 "offen", so data can be inserted on
        # project and year level. This would come in from the master db anyway,
        # but make sure it really has the correct value here. 
        # As uuid may change but the setting has to be active on dtp, its 
        # first being searched on db, then updated on dtp.
        sql = """select DI051826 from DT333
                     where DI051831='p_project_budget_edit'"""
        project_budget_edit_uuid = ppms.db_select(sql)[0][0]
        gs_rec = ppms.search_record(333, [project_budget_edit_uuid],
                                         ['alpha120'])
        gs_rec.alpha120.set_value('0')
        gs_rec.save()
    
    def new_dt280_already_filled(self):
        # check that there is no data in dt480
        count_dt280_data_query = "select count(*) from DT280"
        # check that old data exists
        count_dt495_data_query = "select count(*) from DT495"
        count_dt280_data = self.select_and_log(count_dt280_data_query)[0][0]
        try:
            count_dt495_data = self.select_and_log(count_dt495_data_query)[0][0]
        except ppms.DatabaseError:
            # in voe the table does still exist while in updates from venus the table is already deleted
            # the migration packet needs to only run when the table exists and is filled
            count_dt495_data = 0
        if int(count_dt280_data) == 0 and int(count_dt495_data) > 0:
            return False
        return True
    
    def migrate_cost_type_groups(self):
        self.log_heading('Creating cost type groups')
               
        # migrate the five old cost type groups from dt495 to dt280. The short
        # cost type group id (FACH-ID) didnt exist in Earth and is therefore
        # created here, depending on the user language
        cost_ctg_dict = {1: {'DE':'INT-K', 'EN':'INT-C', 
                             'FR':'C-INT', 'PT':'C-INT'},
                         2: {'DE':'EXT-K','EN':'EXT-C',
                             'FR':'C-EXT','PT':'C-EXT'},
                         3: {'DE':'SACH-K','EN':'MATER-C',
                             'FR':'Coût mat.','PT':'C-MAT'},
                         4: {'DE':'INV-K', 'EN':'INV-C',
                             'FR':'C-INV', 'PT':'C-INV'},
                         5: {'DE':'SONST-K', 'EN':'MISC-C',
                             'FR':'Autr.coût','PT':'OUTR-C'}}
        cur_user_lang = ppms.uvar_get('@19')
        self.new_ctg_ids = {}
        for ctg_id in [1,2,3,4,5]:
            ctg_func_id_dict = cost_ctg_dict[ctg_id]
            ctg_func_id = ctg_func_id_dict[cur_user_lang]
            get_old_ctg = """select DI028496, DI028497,
                                    DI028498, DI028499, DI028500, DI028501, 
                                    DI028502, DI028503, DI028504
                            from DT495 where DI029477 = '{ctg_id}'
                            """.format(**locals())
            old_ctg_data = self.select_and_log(get_old_ctg)[0]
            
            self.log(old_ctg_data)
            todays_date = ppms.uvar_get('@15')
            new_ctg_rec = ppms.create_record(280,
                                {'created_on':todays_date,
                                 'modified_on':todays_date,
                                 'rwdu':old_ctg_data[0],
                                 'created_by':'P20',
                                 'modified_by':'P20',
                                 'own_lic':old_ctg_data[3],
                                 'opc':old_ctg_data[4],
                                 'owner':old_ctg_data[5],
                                 'group':old_ctg_data[6],
                                 'any':old_ctg_data[7],
                                 'ctg_name':old_ctg_data[8],
                                 'cost_type_group':ctg_func_id,
                                 'ctg_description':old_ctg_data[8],
                                 'ctg_active':1,
                                 'ctg_sorting':ctg_id,
                                 'ctg_type_id':306})
            new_ctg_rec.save()
            if not new_ctg_rec:
                self.log('could not create cost type group {ctg_id} ' \
                         '{ctg_func_id}'.format(**locals()))
                return False
            
            new_ctg_id = new_ctg_rec.cost_type_group_id.get_value()
            self.new_ctg_ids[ctg_id] = new_ctg_id
            rec_state = new_ctg_rec.state()
            self.log('created new cost type group {ctg_func_id} tech_id: ' \
                     '{new_ctg_id}, rec_state: {rec_state}'.format(**locals()))  
        
        # create new effort cost type groups
        sorting = 6
        effort_fach_ids = {6: {'DE':'A-INTERN','EN':'E-INTERNAL',
                               'FR':'CH-INTERNE','PT':'E-INTERNO'},
                           7: {'DE':'A-EXTERN','EN':'E-EXTERNAL',
                               'FR':'CH-EXTERNE','PT':'E-EXTERNO'},
                           8: {'DE':'A-SONST','EN':'E-MISCEL',
                               'FR':'CH-AUTRE','PT':'E-OUTRO'}}
        
        effort_names = {6: {'DE':'Interner Aufwand','EN':'Internal effort',
                            'FR':'Charge interne','PT':'Esforço interno'},
                       7: {'DE':'Externer Aufwand','EN':'External effort',
                           'FR':'Charge externe', 'PT':'Esforço externo'},
                       8: {'DE':'Sonstiger Aufwand','EN':'Miscellaneous effort',
                           'FR':'Autre charge','PT':'Outro esforço'}}
                                
        todays_date = ppms.uvar_get('@15')
        cur_user = self.cur_user
        for ctg_func_id in [6,7,8]:
            ctg_info = effort_fach_ids[ctg_func_id]
            ctg_func_fach_id = ctg_info[cur_user_lang]
            ctg_func_name = effort_names[ctg_func_id][cur_user_lang]
            new_ctg_rec = ppms.create_record(280,
                            {'created_on':todays_date,
                             'modified_on':todays_date,
                             'rwdu':7,
                             'created_by':cur_user,
                             'modified_by':cur_user,
                             'own_lic':ppms.uvar_get('@3'),
                             'owner':7, 'group':7, 'any':7,
                             'ctg_name':ctg_func_name,
                             'cost_type_group':ctg_func_fach_id,
                             'ctg_description':ctg_func_name,
                             'ctg_active':1, 
                             'ctg_sorting':sorting,
                             'ctg_type_id':307})
            new_ctg_rec.save()
            if not new_ctg_rec:
                self.log('could not create cost type group {ctg_func_id}, '\
                         '{ctg_func_name}'.format(**locals()))
                return False
                
            new_ctg_id = new_ctg_rec.cost_type_group_id.get_value()
            
            # save new ctg id to 6 for internal and 7 for external effort
            self.new_ctg_ids[sorting] = new_ctg_id
            
            rec_state = new_ctg_rec.state()
            self.log('created new cost type group {ctg_func_fach_id} tech_id: ' \
                     '{new_ctg_id}, rec_state: {rec_state}'.format(**locals()))
            sorting += 1 
              
        # migrate KC->KoAGr-assignments in DT470
        self.modify_and_log(self.set_ctg_to_kc_assignments)
                        
        # save new ctg ids for following migrations
        self.log(self.new_ctg_ids)
        
        self.ctg_id_int_k = self.new_ctg_ids[1]
        self.ctg_id_ext_k = self.new_ctg_ids[2]
        self.ctg_id_sach_k = self.new_ctg_ids[3]
        self.ctg_id_inv_k = self.new_ctg_ids[4]
        self.ctg_id_sonst_k = self.new_ctg_ids[5]
    
        self.ctg_id_a_intern = self.new_ctg_ids[6]
        self.ctg_id_a_extern = self.new_ctg_ids[7]
        self.ctg_id_a_sonst = self.new_ctg_ids[8]
        
        # set default value for effort cost type group assignments on KCs
        self.modify_and_log(self.set_effort_ctg_default.format(**self.__dict__))
        ppms.db_commit()
        return True
    
    def execute_mig_statement(self, ctg_info_dict, mig_statement):
        for ctg_id, source_columns in ctg_info_dict.items():
            dbms_timestamp = self.dbms_timestamp
            self.modify_and_log(mig_statement.format(**locals()))            

    def migrate_benefit_and_revenues(self):
        """ Migration of benefits and revenues in DT835. These values
         have been unstructured before (without subprojects). As we got
         structured and unstructured values in Venus, the migration writes
         the unstructured values to the new DIs and then does a aggregation
         over subprojects to the old DIs which are the structured ones now.
        """
        # copy unstructured benefits+revenues to new columns
        if self.dbms_name == DBMS_ORACLE:
            self.modify_and_log(self.copy_benefit_revenues)
        else:
            self.modify_and_log(self.copy_benefit_revenues_mssql)

        # aggregate unstructured values into old columns
        for pr_id, report_dict in self.all_report_pr_dict.items():
            for report_id, sub_pr_list in report_dict.items():
                if self.dbms_name == DBMS_ORACLE:
                    self.modify_and_log(self.calc_benefit_revenues.format(**locals()))
                else:
                    self.modify_and_log(self.calc_benefit_revenues_mssql.format(**locals()))

        # initialize new column "resulting benefit"
        self.modify_and_log(self.calc_resulting_benefit)

    def migrate_report_years(self):
        """ Migration of data per report and year (DT835 -> DT285) of: 
        *costs planned/actual/remaining
        *budget requested/approved
         
        *effort planned/actual/remaining
        *effort budget requested/approved         
         
        First step migrates current project data to the equivalent 39.5 columns. 
        Second step does a structured aggregation to get total values.
        """
         
        # Step1: migrate earth to venus columns
        ctg_info_dict = {}
        ctg_info_dict[self.ctg_id_int_k] = """
                                  DI029288, -- pl_int_personnel
                                  DI029289, -- act_int_personnel
                                  DI029290, -- rem_int_personnel
                                  DI029291, -- total_int_personnel
                                  DI042769, -- person_int_pay_obl
                                  DI029287, -- int_personnel_budget
                                  DI042774  -- requested_person_int_budget
                                  """
            
        ctg_info_dict[self.ctg_id_ext_k] = """
                                  DI029293, -- pl_ext_personnel
                                  DI029294, -- act_ext_personnel
                                  DI029295, -- rem_ext_personnel
                                  DI029296, -- total_ext_personnel
                                  DI042770, -- person_ext_pay_obl
                                  DI029292, -- ext_personnel_budget
                                  DI042776  -- requested_person_ext_budget
                                  """
                                  
        ctg_info_dict[self.ctg_id_sach_k] = """
                                  DI029298, -- pl_material_costs
                                  DI029299, -- act_material_costs
                                  DI029300, -- rem_material_costs
                                  DI029301, -- total_material_costs
                                  DI042771, -- mat_costs_pay_obl
                                  DI029297, -- material_costs_bdgt
                                  DI042778  -- requested_mat_costs_budget
                                  """
                                   
        ctg_info_dict[self.ctg_id_inv_k] = """
                                  DI029303, -- plan_investments
                                  DI029304, -- actual_investments
                                  DI029305, -- rem_investments
                                  DI029306, -- total_investments
                                  DI042772, -- investm_pay_oblig
                                  DI029302, -- investments_budget
                                  DI042780  -- requested_investm_budget
                                  """
        
        ctg_info_dict[self.ctg_id_sonst_k] = """
                                  DI029308, -- misc_budget
                                  DI029309, -- actual_miscellaneous
                                  DI029310, -- rem_miscellaneous
                                  DI029311, -- total_miscellaneous
                                  DI042773, -- miscel_pay_oblig
                                  DI029307, -- misc_budget
                                  DI042782  -- requested_miscel_budget
                                  """
        
        self.execute_mig_statement(ctg_info_dict, self.mig_report_costs_year)

        """ Efforts """
        ctg_info_dict = {}
        ctg_info_dict[self.ctg_id_a_intern] = """
                                  DI042785, -- effort_int_plan
                                  DI029257, -- act_effort_internal
                                  DI029259, -- rem_effort_internal
                                  DI042786, -- effort_int_tot
                                  DI029255, -- reqstd_effort_int
                                  DI029261  -- appr_effort_internal
                                  """
                                  
        ctg_info_dict[self.ctg_id_a_extern] = """
                                  DI042787, -- effort_ext_plan
                                  DI029258, -- act_effort_external
                                  DI029260, -- rem_effort_external
                                  DI042788, -- effort_ext_tot
                                  DI029256, -- reqstd_effort_ext
                                  DI029262  -- appr_effort_external
                                  """
                                  
        self.execute_mig_statement(ctg_info_dict, self.mig_report_eff_year)

        """ Create year records in the parent project which exist in one of
        it's sub projects but not in the parent. This is needed for
        structured aggregation of cost/efforts afterwards."""
        pr_list = ppms.db_select("""select DI008413, DI008416
                                    from DT823
                                    group by DI008413, DI008416""")
        active_ctg_info = self.get_active_ctgs()
        dbms_timestamp = self.dbms_timestamp
        own_lic = self.usr_license
        for pr_id, report_dict in self.all_report_pr_dict.items():
            for report_id, sub_pr_list in report_dict.items():
                # get years of the sub projects which don't exist in the parent
                get_sub_pr_years = """select distinct DI029231
                                  from DT835
                                  where DI029230 in {sub_pr_list}
                                  and DI029327='{report_id}'
                                  and DI029231 not in (select DI029231
                                                       from DT835
                                                       where DI029230='{pr_id}'
                                                       and DI029327='{report_id}')
                                  """.format(**locals())
                sub_pr_years = ppms.db_select(get_sub_pr_years)
                if not sub_pr_years:
                    continue

                # create year records
                for year in sub_pr_years:
                    self.log('Creating year {year} for project {pr_id} as it ' \
                            'exists in a subproject'.format(**locals()))
                    year = int(year[0])

                    self.modify_and_log(self.mig_report_costs_year_835_subpr.format(**locals()))

                    for ctg_id, ctg_info in active_ctg_info.items():
                        ctg_group = ctg_info['cost_type_group']
                        ctg_name = ctg_info['ctg_name']
                        ctg_sorting = ctg_info['ctg_sorting']
                        ctg_type_code = ctg_info['ctg_type_code']

                        self.modify_and_log(self.mig_report_costs_year_285_subpr.format(**locals()))

        # Step2: aggregate subproject data within new columns to totals
        pr_list = ppms.db_select("select distinct DI060386 from DT285")
        for pr_id, report_dict in self.all_report_pr_dict.items():
            for report_id, sub_pr_list in report_dict.items():
                if self.dbms_name == DBMS_ORACLE:
                    set_year_totals = self.calc_report_year_totals
                else:
                    set_year_totals = self.calc_report_year_totals_mssql

                self.modify_and_log(set_year_totals.format(**locals()))

        # migrate "last baseline" values from new total cost/effort values
        # of active baseline
        if self.dbms_name == DBMS_ORACLE:
            self.modify_and_log(self.calc_bl_budget_report_year)
        else:
            self.modify_and_log(self.calc_bl_budget_report_year_mssql)

    def migrate_report_project(self):
        """ Costs """
        ctg_info_dict = {}
        ctg_info_dict[self.ctg_id_int_k] = """
                            DI028509, -- int_personnel_budget
                            DI028510, -- pl_int_personnel
                            DI028511, -- act_int_personnel
                            DI028512, -- rem_int_personnel
                            DI028513, -- total_int_personnel
                            DI041710  -- personnel_int_pay_oblig"""

        ctg_info_dict[self.ctg_id_ext_k] = """
                            DI028514, -- ext_personnel_budget
                            DI028515, -- pl_ext_personnel
                            DI028516, -- act_ext_personnel
                            DI028517, -- rem_ext_personnel
                            DI028518, -- total_ext_personnel,
                            DI041711  -- personnel_ext_pay_oblig"""
        
        ctg_info_dict[self.ctg_id_sach_k] = """                    
                            DI028519, -- material_costs_bdgt
                            DI028520, -- pl_material_costs
                            DI028521, -- act_material_costs
                            DI028522, -- rem_material_costs
                            DI028523, -- total_material_costs
                            DI041712  -- mat_costs_pay_obl"""
                            
        ctg_info_dict[self.ctg_id_inv_k] = """
                            DI028524, -- investments_budget
                            DI028525, -- plan_investments
                            DI028526, -- actual_investments
                            DI028527, -- rem_investments
                            DI028528, -- total_investments
                            DI041713  -- miscel_pay_oblig"""
                            
        ctg_info_dict[self.ctg_id_sonst_k] = """
                            DI028529, -- misc_budget
                            DI028530, -- planned_misc
                            DI028531, -- actual_miscellaneous
                            DI028532, -- rem_miscellaneous
                            DI028533, -- total_miscellaneous
                            DI041714  -- total_pay_oblig"""
        
        self.execute_mig_statement(ctg_info_dict, self.mig_report_costs_pr)

        # Calculate costs per project/report in DT284 without subprojects,
        # as these existed in Earth with subprojects only
        if self.dbms_name == DBMS_ORACLE:
            self.modify_and_log(self.calc_cost_totals_report)
        else:
            self.modify_and_log(self.calc_cost_totals_report_mssql)

        # Calculate submitted and approved budget values as sum of the
        # year values. These values didn't exist in Earth so they can't be
        # migrated directly
        ctg_recalc_dict = {self.ctg_id_int_k:
                           """sum(nvl(DI042774,0)) requested_person_int_budget,
                              sum(nvl(decode(DI029268, 1, DI029287, 0),
                                      0)) int_personnel_budget""",
                           self.ctg_id_ext_k:
                           """sum(nvl(DI042776,0)) requested_person_ext_budget,
                              sum(nvl(decode(DI029268, 1, DI029292, 0),
                                      0)) ext_personnel_budget""",
                           self.ctg_id_sach_k:
                           """sum(nvl(DI042778,0)) requested_mat_costs_budget,
                              sum(nvl(decode(DI029268, 1, DI029297, 0),
                                      0)) material_costs_bdgt""",
                           self.ctg_id_inv_k:
                           """sum(nvl(DI042780,0)) investments_budget,
                              sum(nvl(decode(DI029268, 1, DI029302, 0),
                                      0)) requested_investm_budget""",
                           self.ctg_id_sonst_k:
                           """sum(nvl(DI042782,0)) requested_miscel_budget,
                              sum(nvl(decode(DI029268, 1, DI029307, 0),
                                      0)) misc_budget"""}

        ctg_recalc_dict_mssql = {self.ctg_id_int_k:
                                 """sum(isnull(DI042774,0)) requested_budget,
                                    sum(case when DI029268=1
                                        then isnull(DI029287, 0)
                                        else 0
                                        end) approved_budget""",
                                 self.ctg_id_ext_k:
                                 """sum(isnull(DI042776,0)) requested_budget,
                                    sum(case when DI029268=1
                                        then isnull(DI029292, 0)
                                        else 0
                                        end) approved_budget""",
                                 self.ctg_id_sach_k:
                                 """sum(isnull(DI042778,0)) requested_budget,
                                    sum(case when DI029268=1
                                        then isnull(DI029297, 0)
                                        else 0
                                        end) approved_budget""",
                                 self.ctg_id_inv_k:
                                 """sum(isnull(DI042780,0)) requested_budget,
                                    sum(case when DI029268=1
                                        then isnull(DI029302, 0)
                                        else 0
                                        end) approved_budget""",
                                 self.ctg_id_sonst_k:
                                 """sum(isnull(DI042782,0)) requested_budget,
                                    sum(case when DI029268=1
                                        then isnull(DI029307, 0)
                                        else 0
                                        end) approved_budget"""}

        if self.dbms_name == DBMS_ORACLE:
            self.execute_mig_statement(ctg_recalc_dict,
                                       self.calc_req_appr_budgets_costs)
        else:
            self.execute_mig_statement(ctg_recalc_dict_mssql,
                                       self.calc_req_appr_budgets_costs_mssql)

        # Aggregate submitted and approved budget values over subprojects
        for pr_id, report_dict in self.all_report_pr_dict.items():
            for report_id, sub_pr_list in report_dict.items():
                if self.dbms_name == DBMS_ORACLE:
                    self.modify_and_log(self.calc_req_appr_budgets_totals_costs.format(**locals()))
                else:
                    self.modify_and_log(self.calc_req_appr_budgets_totals_costs_mssql.format(**locals()))

        """ Efforts """
        # aggregate year values to project (current and total values)
        cur_user = self.cur_user
        all_reports = ppms.db_select("select DI008413, DI008416 from DT823")
        dbms_timestamp = self.dbms_timestamp 
        usr_license = self.usr_license
        effort_groups = (self.ctg_id_a_intern, self.ctg_id_a_extern,
                         self.ctg_id_a_sonst)
        for report_info in all_reports:
            pr_id = report_info[0]
            report_id = report_info[1]

            get_years = """select count(*)
                            from DT285
                            where DI060386='{pr_id}'
                              and DI060387='{report_id}'
                              and DI060389 in {effort_groups}
                            """.format(**locals())
            year_amount = self.select_and_log(get_years)
            if year_amount[0][0] > 0:
                # years exist -> create DT284 rec and aggregate costs from years
                self.modify_and_log(self.mig_report_effort_pr.format(**locals()))
            else:
                # years don't exist -> create empty DT284 records
                for ctg_id in effort_groups:
                    self.modify_and_log(self.mig_report_effort_pr_empty_rec.format(**locals()))

        # migrate effort budget total (ext+int) to DT284 effort budget int
        if self.dbms_name == DBMS_ORACLE:
            self.modify_and_log(self.mig_report_current_budget.format(**self.__dict__))
        else:
            self.modify_and_log(self.mig_report_current_budget_mssql.format(**self.__dict__))
        
        """ Costs and Effort: Aggregation of former virtual DIs.
        - get list of all projects which have a status report and 
        - aggregate budget total as sum of actual budget over subprojects
        - this runs seperated to aggregation above, because budget/year didn't
        exist in earth."""

        for pr_id, report_dict in self.all_report_pr_dict.items():
            for report_id, sub_pr_list in report_dict.items():
                if self.dbms_name == DBMS_ORACLE:
                    self.modify_and_log(self.calc_budget_total_report.format(**locals()))
                else:
                    self.modify_and_log(self.calc_budget_total_report_mssql.format(**locals()))
            
        # Calculate submitted and approved budget values as sum of the
        # year values. These values didn't exist in VoE so they can't be
        # migrated directly
        ctg_recalc_dict_effort = {self.ctg_id_a_intern:
                           """sum(nvl(DI029255,0)) reqstd_effort_int,
                              sum(nvl(decode(DI029268, 1, DI029261, 0),
                                      0)) appr_effort_internal""",
                           self.ctg_id_a_extern:
                           """sum(nvl(DI029256,0)) reqstd_effort_ext,
                              sum(nvl(decode(DI029268, 1, DI029262, 0),
                                     0)) appr_effort_external"""}

        ctg_recalc_dict_effort_mssql = {self.ctg_id_a_intern:
                                 """sum(isnull(DI029255,0)) budget_submitted_eff,
                                    sum(case when DI029268=1
                                        then isnull(DI029261, 0)
                                        else 0
                                        end) budget_approved_eff""",
                                 self.ctg_id_a_extern:
                                 """sum(isnull(DI029256,0)) budget_submitted_eff,
                                    sum(case when DI029268=1
                                        then isnull(DI029262, 0)
                                        else 0
                                        end) budget_approved_eff"""}

        all_projects = ppms.db_select("select distinct DI008413 from DT823")
        if self.dbms_name == DBMS_ORACLE:
            self.execute_mig_statement(ctg_recalc_dict_effort,
                                       self.calc_req_appr_budgets_effort)
        else:
            self.execute_mig_statement(ctg_recalc_dict_effort_mssql,
                                       self.calc_req_appr_budgets_effort_mssql)

        # Aggregate submitted and approved budget values over subprojects
        for pr_id, report_dict in self.all_report_pr_dict.items():
            for report_id, sub_pr_list in report_dict.items():
                if self.dbms_name == DBMS_ORACLE:
                    self.modify_and_log(self.calc_req_appr_budgets_totals_effort.format(**locals()))
                else:
                    self.modify_and_log(self.calc_req_appr_budgets_totals_effort_mssql.format(**locals()))

        #Calculate effort and cost budget of subprojects as sum of budget values of subprojects without current project
        for pr_id, report_dict in self.all_report_pr_dict.items():
            for report_id in report_dict.keys():
                sub_pr_list = self.get_pr_tree_desc(pr_id, report_id,
                                                    list_format='db',
                                                    with_main_pr=False)

                #no need of aggregation, when no sub_pr exists
                if len(sub_pr_list) != 0:
                    if self.dbms_name == DBMS_ORACLE:
                        self.modify_and_log(self.calc_budget_subproj_report.format(**locals()))
                    else:
                        self.modify_and_log(self.calc_budget_subproj_report_mssql.format(**locals()))

        # migrate "last baseline" values from new total cost/effort values
        # of active baseline
        if self.dbms_name == DBMS_ORACLE:
            self.modify_and_log(self.calc_bl_budget_report_project)
        else:
            self.modify_and_log(self.calc_bl_budget_report_project_mssql)

    def migrate_budget_years(self):
        """ Cost+Budget / Project/Year/CTG """
        ctg_info_dict = {}
        ctg_info_dict[self.ctg_id_int_k] = """
                               DI041635, -- act_personnel_int_pl
                               DI041636, -- act_personnel_int_act
                               DI041637, -- act_personnel_int_rem
                               DI041639, -- act_personnel_int_tot
                               DI028558, -- int_personnel_budget
                               DI040198, -- appr_int_personnel
                               DI041638  -- act_personnel_int_pay_oblig"""
        
        ctg_info_dict[self.ctg_id_ext_k] = """
                               DI041640, -- act_personnel_ext_pl
                               DI041641, -- act_personnel_ext_act
                               DI041642, -- act_personnel_ext_rem
                               DI041644, -- act_personnel_ext_tot
                               DI028563, -- ext_personnel_budget
                               DI040199, -- appr_ext_personnel
                               DI041643  -- act_personnel_ext_pay_oblig"""
        
        ctg_info_dict[self.ctg_id_sach_k] = """
                               DI041645, -- act_mat_costs_pl
                               DI041646, -- act_mat_costs_act
                               DI041647, -- act_mat_costs_rem
                               DI041649, -- act_mat_costs_tot
                               DI028568, -- material_costs_bdgt
                               DI040200, -- appr_material_costs
                               DI041648  -- act_mat_costs_pay_oblig"""
        
        ctg_info_dict[self.ctg_id_inv_k] = """
                               DI041650, -- act_investm_pl
                               DI041651, -- act_investm_act
                               DI041652, -- act_investm_rem
                               DI041654, -- act_investm_tot 
                               DI028573, -- investments_budget
                               DI040201, -- appr_investments
                               DI041653  -- act_investm_pay_oblig"""
                               
        ctg_info_dict[self.ctg_id_sonst_k] = """
                               DI041655, -- act_miscellaneous_pl
                               DI041656, -- act_miscellaneous_act
                               DI041657, -- act_miscellaneous_rem
                               DI041659, -- act_miscellaneous_tot
                               DI028578, -- misc_budget
                               DI040202, -- appr_miscellaneous                               
                               DI041658  -- act_miscellaneous_pay_oblig"""
        
        self.execute_mig_statement(ctg_info_dict, self.mig_costs_pr_year)  

        """ Effort """
        ctg_info_dict= {}
        ctg_info_dict[self.ctg_id_a_intern] = """
                               DI041627, -- act_effort_int_pl
                               DI041628, -- act_effort_int_act
                               DI041629, -- act_effort_int_rem
                               DI041630, -- act_effort_int_tot
                               DI023165, -- reqstd_effort_int
                               DI023178  -- appr_effort_internal
                               """
                               
        ctg_info_dict[self.ctg_id_a_extern] = """
                               DI041631, -- act_effort_ext_pl
                               DI041632, -- act_effort_ext_act
                               DI041633, -- act_effort_ext_rem
                               DI041634, -- act_effort_ext_tot
                               DI023166, -- reqstd_effort_ext
                               DI023179  -- appr_effort_external
                               """
        self.execute_mig_statement(ctg_info_dict, self.mig_eff_pr_year)  
        
    def migrate_budget_project(self):
        """ Budget/Project/CTG
        The costs in DT461 are not migrated here - as these will
        be calculated during new-planning anyway. therefore migrating
        budget only."""
        ctg_info_dict_cost = {self.ctg_id_int_k:
                              "DI028309 -- int_personnel_budget",
                              self.ctg_id_ext_k:
                              "DI028318 -- ext_personnel_budget",
                              self.ctg_id_sach_k:
                              "DI028319 -- material_costs_bdgt",
                              self.ctg_id_inv_k:
                              "DI028324 -- investments_budget",
                              self.ctg_id_sonst_k:
                              "DI028329 -- misc_budget"}
                                
        self.execute_mig_statement(ctg_info_dict_cost, self.mig_costs_pr)
        
        # Migrate effort "gesamt" to internal effort, as there was no cost 
        # type group distribution before venus and the "gesamt" value is
        # virtual in venus 
        ctg_info_dict_effort = {self.ctg_id_a_intern:
                                "DI006926 -- effort_budget"}
        
        self.execute_mig_statement(ctg_info_dict_effort, self.mig_effort_pr)
                                
    def migrate_portfolio_budget(self):
        """ Migrates portfolio budget per cost type group from DT859 to DT283
        """
        self.log('Starting budget migration of all portfolios')
        pf_statements = {DBMS_ORACLE: [self.mig_pf_budget_int, 
                                       self.mig_pf_budget_ext, 
                                       self.mig_pf_budget_sach, 
                                       self.mig_pf_budget_inv, 
                                       self.mig_pf_budget_sonst,
                                       self.mig_pf_eff_budget_int,
                                       self.mig_pf_eff_budget_ext],
                         DBMS_MSSQL:  [self.mig_pf_budget_int_mssql,
                                       self.mig_pf_budget_ext_mssql,
                                       self.mig_pf_budget_sach_mssql,
                                       self.mig_pf_budget_inv_mssql, 
                                       self.mig_pf_budget_sonst_mssql,
                                       self.mig_pf_eff_budget_int_mssql,
                                       self.mig_pf_eff_budget_ext_mssql]}
        
        for stmt in pf_statements[self.dbms_name]:
            self.modify_and_log(stmt.format(**self.__dict__))
        
        self.log('Portfolio budget migration done.')

    def activate_pr_year_pf(self):
        """ Activate all years in DT286 for all projects which are assigned to 
        a portfolio """
        self.log_heading('Activating portfolio year scales in DT286')
        self.modify_and_log(self.activate_pf_years.format(**self.__dict__))
        
    def get_pr_tree_desc(self, proj_id, report_id=None,
                         list_format='py', with_main_pr=True):
        """ Returns a descending list of child projects of a project.
        - If with_main_pr=False without proj_id itself
        - If a report_id is set -> get the structure from DT823
          If no report_id is set -> get the structure from DT461"""

        if report_id:
            pr_select = "Select DI008413 From DT823 " \
                        "where DI008414 = '{0}' and DI008416 = '{1}'"
        else:
            pr_select = "Select DI001001 From Dt461 Where di001003 = '{0}'"

        subprojects = ppms.db_select(pr_select.format(proj_id, report_id))
        scanned_projects = []
        if with_main_pr:
            scanned_projects.append(proj_id)
        for project in subprojects:
            if report_id:
                sub_list = ppms.db_select(pr_select.format(project[0], report_id))
            else:
                sub_list = ppms.db_select(pr_select.format(project[0]))

            for next_subproj in sub_list:
                if next_subproj not in subprojects:
                    subprojects.append(next_subproj)
            scanned_projects.append(project[0])

        if list_format == 'db':
            # return tuple-like str to use it in sql select
            scanned_projects = tuple(scanned_projects)
            if len(scanned_projects) == 1:
                scanned_projects = (scanned_projects[0], scanned_projects[0])
        return scanned_projects

    def get_active_ctgs(self):
        # get info of all active ctg
        usr_language = ppms.uvar_get('@19')
        ctg_qry = """select
                        DI060121, DI060132,DI060133,
                        (select DI001549
                        from DT435
                        where DI001538=DI041277 and DI001550 = '{0}'),
                        (select DI001549
                        from DT435
                        where DI001538=DI041276 and DI001550 = '{0}'),
                        DI060136,DI060135
                    from DT280,DT447
                    where DT280.DI060467 = DT447.DI041264
                    and DI060135 = 1""".format(usr_language)
        ctg_info_280 = self.select_and_log(ctg_qry)
        active_ctgs = {}
        for ctg_info in ctg_info_280:
            ctg_id = str(ctg_info[0])
            active_ctgs[ctg_id] = {'obj_id':'',
                                   'cost_type_group_id':ctg_id,
                                   'ctg_name':ctg_info[1],
                                   'cost_type_group':ctg_info[2],
                                   'ctg_type_code':ctg_info[3],
                                   'ctg_type_name':ctg_info[4],
                                   'ctg_sorting':int(ctg_info[5])}
        return active_ctgs

    def invoke_update_284(self):
        # Create new record in DT284 (project/report/cost type group) if it 
        # doesn't exists yet
        ctg_check = """select 1 
                         from DT284
                         where DI060297 = '{obj_id}' 
                         and DI060299 = '{ctg_id}'
                         and DI060300 = '{report_id}'""".format(**self.__dict__)
        ctg_exists = self.select_and_log(ctg_check)
        if ctg_exists:
            return
                        
        # fill statement with ctg info               
        update_target_dt = self.create_report_pr.format(**self.__dict__)
        ppms.db_modify(update_target_dt)
        self.count_update += 1

    def invoke_update_285(self):
        # Create new record in DT285 (project/report/year/cost type group)
        # if it doesn't exists yet
        year_qry = """select DI029231
                      from DT835 
                      where DI029230 = '{obj_id}'
                      and DI029327 = '{report_id}'""".format(**self.__dict__)
        existing_years = ppms.db_select(year_qry)
        
        # create 285 records if there are existing year scales in dt861
        target_check = """select 1 
                            from DT285 
                            where DI060386 = '{obj_id}'
                            and DI060388 = {year}
                            and DI060389 = '{ctg_id}'
                            and DI060387 = '{report_id}'"""

        for year in existing_years:
            if isinstance(year, list):
                year = year[0]
            self.year = int(year)
            check_qry = target_check.format(**self.__dict__)
            
            ctg_rec_exists = ppms.db_select(check_qry)
            if ctg_rec_exists:
                continue
            
            # create new record if not existant    
            update_qry = self.create_report_pr_year.format(**self.__dict__)
            self.modify_and_log(update_qry)
            self.count_update += 1

    def start_report_ctg_transfer(self):
        # transfer project and ctg data into dt284+285
        self.count_update = 0    
        reports = self.select_and_log("""select DI008413, DI008416 
                                          from dt823""")
        self.active_ctgs = self.get_active_ctgs()
        
        for ctg in self.active_ctgs:
            self.ctg_info = self.active_ctgs[ctg]
            self.ctg_id = self.ctg_info['cost_type_group_id']
            self.ctg_name = self.ctg_info['ctg_name']
            self.cost_type_group = self.ctg_info['cost_type_group']
            self.ctg_type_code = self.ctg_info['ctg_type_code']
            self.ctg_type_name = self.ctg_info['ctg_type_name']
            self.ctg_sorting = self.ctg_info['ctg_sorting']
            
            for report_info in reports:
                self.obj_id = str(report_info[0])
                self.report_id = str(report_info[1])
                
                self.invoke_update_284()
                self.invoke_update_285()
        
        final_txt = "Cost type group replication done. " \
                    "{count_update} records were modified."
        return final_txt.format(**self.__dict__)
        
